"""
Console Scripts
---------------

To see available commands
::

    wads --help

"""

import setuptools  # to avoid a warning due to distutils_patch.py
import os
import json

root_dir = os.path.dirname(__file__)
root_dir_name = os.path.basename(root_dir)
rjoin = lambda *paths: os.path.join(root_dir, *paths)

data_dir = rjoin("data")
licenses_json_path = rjoin(data_dir, "github_licenses.json")
setup_tpl_path = rjoin(data_dir, "setup_tpl")
gitignore_tpl_path = rjoin(data_dir, ".gitignore_tpl")
github_ci_tpl_path = rjoin(data_dir, "github_ci_tpl.yml")
gitlab_ci_tpl_path = rjoin(data_dir, "gitlab_ci_tpl.yml")
github_ci_tpl_deploy_path = rjoin(data_dir, "github_ci_tpl_deploy.yml")
github_ci_tpl_publish_path = rjoin(data_dir, "github_ci_tpl_publish.yml")

pkg_dir = os.path.dirname(root_dir)
pkg_join = lambda *paths: os.path.join(pkg_dir, *paths)

# TODO: Change to use ini format? (Or yaml or toml?)
wads_configs_file = rjoin(data_dir, "wads_configs.json")
try:
    wads_configs = json.load(open(wads_configs_file))
except FileNotFoundError:
    wads_configs = {
        "populate_dflts": {
            "description": "There is a bit of an air of mystery around this project...",
            "root_url": None,
            "author": None,
            "license": "mit",
            "description_file": "README.md",
            "long_description": "file:README.md",
            "long_description_content_type": "text/markdown",
            "keywords": None,
            "install_requires": None,
            "verbose": None,
            "version": "0.0.1",
            "project_type": "lib",
        }
    }

pkg_path_names = (".gitignore", "setup.py")

from wads.populate import populate_pkg_dir


def main():
    import argh  # pip install argh

    from wads.pack import argh_kwargs as pack_kw

    # from wads.docs_gen import argh_kwargs as docs_gen_kw

    parser = argh.ArghParser()
    parser.add_commands(**pack_kw)
    # parser.add_commands(**docs_gen_kw)
    parser.dispatch()


if __name__ == "__main__":
    main()



# TODO: notebook -> md with more parametrization, normalization, and cleaning up

# TODO: Get rid of html junk in ipynb -> md

# TODO: Transform markdown (md) relative image reference to full url ones so that they're rendered in pypi
# Example: ![png](img/output_72_1.png) -> ![png](https://github.com/thorwhalen/invest/tree/master/img/output_72_1.png)

# https://raw.githubusercontent.com/thorwhalen/invest/master/img


# TODO: Make it work with regular expressions, to be more flexible
def replace_relative_links_to_absolute(
    s: str,
    root_url: str,
    str_to_match_and_add_root_url_before: str = "\n![png](",
):
    return s.replace(
        str_to_match_and_add_root_url_before,
        root_url + str_to_match_and_add_root_url_before,
    )



import re
from itertools import chain
import pkgutil
from warnings import warn
from functools import partial
import os
import subprocess

standard_lib_dir = os.path.dirname(os.__file__)
path_sep = os.path.sep


def ensure_slash_suffix(s: str):
    if not s.endswith(path_sep):
        s += path_sep
    return s


def ensure_no_slash_suffix(s: str):
    return s.rstrip(path_sep)


def _build_git_command(command: str = "status", work_tree=".", git_dir=None):
    if command.startswith("git "):
        warn(
            "You don't need to start your command with 'git '. I know it's a git command. Removing that prefix"
        )
        command = command[len("git ") :]
    work_tree = os.path.abspath(os.path.expanduser(work_tree))
    if git_dir is None:
        git_dir = os.path.join(work_tree, ".git")
    assert os.path.isdir(git_dir), f"Didn't find the git_dir: {git_dir}"
    git_dir = ensure_no_slash_suffix(git_dir)
    if not git_dir.endswith(".git"):
        warn(f"git_dir doesn't end with `.git`: {git_dir}")
    return f'git --git-dir="{git_dir}" --work-tree="{work_tree}" {command}'


def git(command: str = "status", work_tree=".", git_dir=None):
    """Launch git commands.

    :param command: git command (e.g. 'status', 'branch', 'commit -m "blah"', 'push', etc.)
    :param work_tree: The work_tree directory (i.e. where the project is)
    :param git_dir: The .git directory (usually, and by default, will be taken to be "{work_tree}/.git/"
    :return: What ever the command line returns (decoded to string)
    """

    """

    git --git-dir=/path/to/my/directory/.git/ --work-tree=/path/to/my/directory/ add myFile
    git --git-dir=/path/to/my/directory/.git/ --work-tree=/path/to/my/directory/ commit -m 'something'

    """
    command_str = _build_git_command(command, work_tree, git_dir)
    r = subprocess.check_output(command_str, shell=True)
    if isinstance(r, bytes):
        r = r.decode()
    return r.strip()


def is_standard_lib_path(path):
    return path.startswith(standard_lib_dir)


def standard_lib_module_names(
    is_standard_lib_path=is_standard_lib_path,
    name_filt=lambda name: not name.startswith("_"),
):
    return filter(
        name_filt,
        (
            module_info.name
            for module_info in pkgutil.iter_modules()
            if is_standard_lib_path(module_info.module_finder.path)
        ),
    )


def mk_conditional_logger(condition, func=print):
    if condition:
        return func
    else:

        def do_nothing(*args, **kwargs):
            pass

        return do_nothing


def clog(condition, func=print, *args, **kwargs):
    if condition:
        func(*args, **kwargs)


def mk_replacer_from_dict(from_to_dict):
    """Make a function that does multiple replacements (in a single pass).

    >>> r = mk_replacer_from_dict({'is': 'are', 'life': 'butterflies'})
    >>> r("There is no life in the void.")
    'There are no butterflies in the void.'

    :param from_to_dict: A dict of {to_find: to_replace_by,...} pairs
    :return: A replacer function that you can apply to strings to carry out the replacements
    """

    p = re.compile("|".join(map(r"({})".format, map(re.escape, from_to_dict.keys()))))
    f = lambda x: from_to_dict[x.group(0)]

    def replacer(s):
        return p.sub(f, s)

    return replacer


def mk_import_root_replacer(from_to_dict):
    """Make a function that does multiple import name replacements.

    For a use case, see replace_import_names. This is just a helper function.

    >>> replace = mk_import_root_replacer({'foo': 'FOO', 'bar': 'BAR'})
    >>>
    >>> assert replace('from foo import BLAH') == 'from FOO import BLAH'
    >>> assert replace('from foo.bar import BLAH') == 'from FOO.bar import BLAH'
    >>> assert replace('import bar') == 'import BAR'

    Partial matches are not replaced (that's a good thing!):

    >>> assert replace('import barmitzvah as oy') == 'import barmitzvah as oy'
    >>> assert replace('from foobar import hello') == 'from foobar import hello'

    Yes, and it works with dotpaths:

    >>> replace = mk_import_root_replacer({'where.it.was': 'where.it.is.now'})
    >>> replace('import where.it.was as here')
    'import where.it.is.now as here'

    :param from_to_dict: A dict of {to_find: to_replace_by,...} pairs
    :return: A replacer function that you can apply to strings to carry out the replacements

    r"""
    t = dict(
        chain.from_iterable(
            [
                (f"(?<=from\ ){old}(?=[\.\ ])", f"{new}"),
                (f"(?<=import\ ){old}(?=[\.\s])", f"{new}"),
            ]
            for old, new in from_to_dict.items()
        )
    )

    p = re.compile("|".join(map(r"({})".format, t.keys())))
    f = lambda x: from_to_dict[x.group(0)]

    def replacer(s):
        ss = p.sub(f, s + " ")
        return ss[:-1]

    return replacer


def replace_import_names(
    source_store,
    from_to_dict,
    target_store=None,
    dryrun=True,
    verbose=True,
    replacer_factory=mk_import_root_replacer,
    add_comment_at_the_end_of_lines_replaced=False,
):
    """Replace import names.

    Use case: You've renamed something or moved some modules (remember UNIX? Same as move!) and have to go through
    all your files and notebooks and replace those names.
    Now, if you have a nice IDE, we suggest you use refactoring instead -- as long as you have any uses in the scope.
    But sometimes it's not enough. You might have text/html documments, or jupyter notebooks, etc.
    So you can use this instead.

    Be warned though:
    - You should look at the pattern that is used to match, and make sure it won't create havoc.
    - Backup your documents so you can revert!
    - Print matches before you actually apply them all in bulk.
    - Use at your own risk!

    For examples, see the mk_import_root_replacer helper function.

    :param source_store:
    :param from_to_dict: A dict of {to_find: to_replace_by,...} pairs
    :param target_store: The store
    :param dryrun:
    :param verbose:
    :param replacer_factory: makes the replacer = replacer_factory(from_to_dict)
    :param add_comment_at_the_end_of_lines_replaced: True/False or an actual string to add at the end of replaced lines
    :return:
    """

    suffix = ""
    if add_comment_at_the_end_of_lines_replaced is False:
        suffix = ""
    elif add_comment_at_the_end_of_lines_replaced is True:
        suffix = "# line_was_edited_by_wads"
    elif isinstance(add_comment_at_the_end_of_lines_replaced, str):
        suffix = add_comment_at_the_end_of_lines_replaced
        if not suffix.startswith("#"):
            suffix = "# " + suffix
    else:
        raise TypeError(
            f"Don't know what to do with such a type of add_comment_at_the_end_of_lines_replaced: "
            f"{add_comment_at_the_end_of_lines_replaced}"
        )

    _clog = mk_conditional_logger(condition=verbose, func=print)

    if target_store is None:
        target_store = {}  # use a dict to write results
        _clog(
            "You didn't specify a target_store, so I'll write all of this in a dict and return it to you!"
        )

    if source_store == target_store:
        # TODO: Add confirmation (user input) to protect more. Make this an option (so total automatic is possible)
        _clog(
            "I just wanted you to be aware that you are using the same store for source and target."
            "This means I'm about to overwrite your files!!"
        )
        if dryrun:
            _clog("... Right now, dryrun=True, so I'm just pretending!")

    replacer = replacer_factory(from_to_dict)

    if dryrun:
        replace_prompt = "will replace"
    else:
        replace_prompt = "   replacing"
    with____prompt = "        with"

    for k, v in source_store.items():
        lines_replaced = 0

        def gen():
            nonlocal lines_replaced
            for i, line in enumerate(
                v.split("\n"), 1
            ):  # TODO: Double traversal. Find online splitter
                new_line = replacer(line) + suffix
                yield new_line
                if (
                    new_line != line
                ):  # TODO: Double traversal... could have replacer return a "replaced_something" flag
                    lines_replaced += 1
                    clog(
                        verbose or dryrun,
                        print,
                        f"{k}:{i}:\n{replace_prompt}\t{line}\n{with____prompt}\t{new_line}",
                    )

        new_v = "\n".join(gen())

        if not dryrun and lines_replaced > 0:
            target_store[k] = new_v

    return target_store


fc = dict(
    reset="\033[0m",  # alias for reset_all
    reset_all="\033[0m",
    bold="\033[1m",
    dim="\033[2m",
    underlined="\033[4m",
    blink="\033[5m",
    reverse="\033[7m",
    hidden="\033[8m",
    reset_bold="\033[21m",
    reset_dim="\033[22m",
    reset_underlined="\033[24m",
    reset_blink="\033[25m",
    reset_reverse="\033[27m",
    reset_hidden="\033[28m",
    default="\033[39m",
    black="\033[30m",
    red="\033[31m",
    green="\033[32m",
    yellow="\033[33m",
    blue="\033[34m",
    magenta="\033[35m",
    cyan="\033[36m",
    gray="\033[37m",
    dark_gray="\033[90m",
    dark_red="\033[91m",
    dark_green="\033[92m",
    dark_yellow="\033[93m",
    dark_blue="\033[94m",
    dark_magenta="\033[95m",
    dark_cyan="\033[96m",
    white="\033[97m",
    background_default="\033[49m",
    background_black="\033[40m",
    background_red="\033[41m",
    background_green="\033[42m",
    background_yellow="\033[43m",
    background_blue="\033[44m",
    background_magenta="\033[45m",
    background_cyan="\033[46m",
    background_gray="\033[47m",
    background_dark_gray="\033[100m",
    background_dark_red="\033[101m",
    background_dark_green="\033[102m",
    background_dark_yellow="\033[103m",
    background_dark_blue="\033[104m",
    background_dark_magenta="\033[105m",
    background_dark_cyan="\033[106m",
    background_white="\033[107m",
)

# Make a namespace out of the fc dict
from types import SimpleNamespace

fcc = SimpleNamespace(**fc)


def highlight(
    string,
    effect=fc["reverse"],
    beg_mark="[[",
    end_mark="]]",
    end_effect=fc["reset_all"],
):
    r"""Interprets a string's highlight markers to be able to make highlights in the string.

    This is meant for very simple situations. A more powerful and fast function could be made by
    using regular expressions and a map to map "codes" to "effects".

    Try this:

    >>> print(highlight("This is [[the section]] that is [[highlighted]]."))  # doctest:+SKIP

    Above, "reverse" is used as the default effect.
    But You can change that to bold blue ink on yellow background. That's three effects:
    1 (for bold), 34, for the blue foreground (ink), and 43 for the "yellow" (more like brown)
    background (paper).

    >>> my_string = "This is [[the section]] that is [[highlighted]]."
    >>> print(highlight(my_string, "\033[1;34;43m"))  # doctest:+SKIP

    \033[whaaaa?!? Yeah... well, either you do it that has-no-life-outside-unicode way.
    If so, Ansi help you!
    See [this wiki section](https://en.wikipedia.org/wiki/ANSI_escape_code#SGR_parameters).
    or [this tutorial](https://www.lihaoyi.com/post/BuildyourownCommandLinewithANSIescapecodes.html#rich-text).

    If not, we've prepared a map between human language and effect codes in the form of the
    `fc` variable of this module. It's a dict (and if you have ``py2store``, it's a mapping containing
    that dict and allowing you access through attributes too).

    >>> from wads.util import fc
    >>> list(fc)[20:25]
    ['magenta', 'cyan', 'gray', 'dark_gray', 'dark_red']

    :param string: String with highlight formatting
    :param effect: The effect to use for the highlighting (some unicode like "\033[21m")
    :param beg_mark: String that marks the beginning of the highlight
    :param end_mark: String that marks the end of the highlight
    :param end_effect: The unicode to use to reset the effect
    :return:
    """
    return string.replace(beg_mark, effect).replace(end_mark, end_effect)


# ------------------------------------------------------------------------------------ #
# A function to import objects from modules when we want to have a backup in case the
# object is not found in the first module.


def find_obj(*module_and_obj_names):
    """Find an object from specified modules, under specified names.

    This function searches for a specified object in a list of specified modules.
    It first tries to import the object from the first module specified, then the
    second, and so on, until it finds the object. If the object is not found in any
    of the specified modules, an ImportError is raised.

    Args:
        *module_and_obj_names: A list of pairs of module names and object names.
            The first module specified is searched first, then the second, and so on,
            until the object is found. Each module and object pair is specified as a
            tuple containing two strings. The first string is the module name, and the
            second string is the object name. Alternatively, the module and object pair
            can be specified as a single string containing two space-separated strings,
            where the first string is the module name and the second string is the
            object name.

    Returns:
        The imported object.

    Raises:
        ImportError: If the object cannot be found in any of the specified modules.

    >>> files = find_obj('importlib.resources files', 'importlib_resources files')
    >>> callable(files)
    True
    >>> find_obj(
    ...     'importlib no_such_obj', 'no_such_pkg x', ['wave.no_such_module', 'y']
    ... ) # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    ImportError: All of these import attempts failed:
        from importlib import no_such_obj
        from no_such_pkg import x
        from wave.no_such_module import y

    See discussion: https://github.com/i2mint/wads/discussions/10.
    """
    module_and_obj_names = list(map(_ensure_pair, module_and_obj_names))
    for module_name, obj_name in module_and_obj_names:
        try:
            return import_obj(module_name, obj_name)
        except (ImportError, ModuleNotFoundError, AttributeError):
            continue
    from_import_statements = "\n\t" + "\n\t".join(
        map("from {} import {}".format, *zip(*module_and_obj_names))
    )
    raise ImportError(f"All of these import attempts failed:{from_import_statements}")


def import_obj(module_name: str, obj_name: str):
    """Import an object from a specified module.

    Args:
        from_ (str): The name of the module to import from.
        obj_name (str): The name of the object to import.

    Returns:
        The imported object.

    Raises:
        ImportError: If the module or object cannot be imported.
        AttributeError: If the object does not exist in the module.

    >>> spec_finder = import_obj('importlib.util', 'find_spec')
    >>> callable(spec_finder)
    True
    >>> spec_finder.__module__
    'importlib.util'

    """
    module = __import__(module_name, fromlist=[obj_name])
    obj = getattr(module, obj_name)
    return obj


def _ensure_pair(x):
    if isinstance(x, str):
        x = x.split()
    assert len(x) == 2, f"should be a pair of strings: {x}"
    return x



import subprocess
import sys
import tarfile
import tempfile
from pathlib import Path

from wads.package_module import generate_package

EXAMPLE_MODULE_FILE = Path(__file__).parent / "data" / "example_module.py"
EXAMPLE_MODULE_DIR = Path(__file__).parent / "data"


def test_generate_package_from_file():
    with tempfile.TemporaryDirectory() as tmp_folder:
        tmp_path = Path(tmp_folder)
        output_path = tmp_path / "output"
        generate_package(
            module_path=EXAMPLE_MODULE_FILE,
            install_requires=["dol", "i2 @ git+https://git@github.com/i2mint/i2.git"],
            output_path=output_path,
            version="8.8.8",
        )
        try:
            subprocess.check_output(
                [sys.executable, "setup.py", "sdist"],
                cwd=output_path,
            )
        except subprocess.CalledProcessError as e:
            error_message = e.output.decode("utf-8")  # Decode the output if needed
            print("Error message:", error_message)
            raise e

        i2_whl = next((output_path / "dist").glob("i2-*.whl"))
        assert i2_whl, "wheel not built from specified git repo"
        module_tar_gz = next((output_path / "dist").glob("*.tar.gz"))
        assert module_tar_gz.is_file()
        with tarfile.open(module_tar_gz, "r:gz") as tar:
            tar.extractall(tmp_path / "unpacked")
        unpacked_dist = (
            tmp_path / "unpacked" / module_tar_gz.name[: -len(".tar.gz")] / "dist"
        )
        unpacked_i2_whl = next(unpacked_dist.glob("i2-*.whl"))
        assert (
            unpacked_i2_whl.read_bytes() == i2_whl.read_bytes()
        ), ".tar.gz does not include dependency wheel"


def test_generate_package_from_dir():
    with tempfile.TemporaryDirectory() as tmp_folder:
        tmp_path = Path(tmp_folder)
        output_path = tmp_path / "output"
        generate_package(
            module_path=EXAMPLE_MODULE_DIR,
            install_requires=["dol", "i2 @ git+https://git@github.com/i2mint/i2.git"],
            output_path=output_path,
            version="8.8.8",
            glob_pattern=["*.pkl", "*.json"],
        )

        try:
            subprocess.check_output(
                [sys.executable, "setup.py", "sdist"],
                cwd=output_path,
            )
        except subprocess.CalledProcessError as e:
            error_message = e.output.decode("utf-8")  # Decode the output if needed
            print("Error message:", error_message)
            raise e
        i2_whl = next((output_path / "dist").glob("i2-*.whl"))
        assert i2_whl, "wheel not built from specified git repo"
        module_tar_gz = next((output_path / "dist").glob("*.tar.gz"))
        assert module_tar_gz.is_file()
        with tarfile.open(module_tar_gz, "r:gz") as tar:
            tar.extractall(tmp_path / "unpacked")
        unpacked_dir = tmp_path / "unpacked" / module_tar_gz.name[: -len(".tar.gz")]
        unpacked_dist = unpacked_dir / "dist"
        unpacked_module = unpacked_dir / EXAMPLE_MODULE_DIR.stem
        unpacked_i2_whl = next(unpacked_dist.glob("i2-*.whl"))
        unpacked_json = next(unpacked_module.glob("*.json"))
        assert unpacked_json.is_file()
        unpacked_pkl = next(unpacked_module.glob("*.pkl"))
        assert unpacked_pkl.is_file()
        assert (
            unpacked_i2_whl.read_bytes() == i2_whl.read_bytes()
        ), ".tar.gz does not include dependency wheel"



import os
import shutil
import json
from collections import ChainMap
from urllib.parse import urlparse

# from functools import partial
from typing import List, Optional
from wads import (
    pkg_join as wads_join,
    gitlab_ci_tpl_path,
    setup_tpl_path,
    gitignore_tpl_path,
    github_ci_tpl_deploy_path,
    github_ci_tpl_publish_path,
    wads_configs,
    wads_configs_file,
)
from wads.util import mk_conditional_logger, git, ensure_no_slash_suffix
from wads.pack import write_configs
from wads.licensing import license_body

# from wads.pack_util import write_configs

path_sep = os.path.sep

populate_dflts = wads_configs.get(
    "populate_dflts",
    {
        "description": "There is a bit of an air of mystery around this project...",
        "root_url": None,
        "author": None,
        "license": "mit",
        "description_file": "README.md",
        "long_description": "file:README.md",
        "long_description_content_type": "text/markdown",
        "keywords": None,
        "install_requires": None,
        "verbose": True,
        "version": "0.0.1",
        "project_type": "lib",
    },
)


def gen_readme_text(
    name, text="There is a bit of an air of mystery around this project..."
):
    return f"""
# {name}
{text}
"""


# TODO: Function way to long -- break it up
# TODO: Add a `defaults_from` in **configs that allows one to have several named defaults in wads_configs_file
def populate_pkg_dir(
    pkg_dir,
    version: str = populate_dflts["version"],
    description: str = populate_dflts["description"],
    root_url: Optional[str] = populate_dflts["root_url"],
    author: Optional[str] = populate_dflts["author"],
    license: str = populate_dflts["license"],
    description_file: str = populate_dflts["description_file"],
    keywords: Optional[List] = populate_dflts["keywords"],
    install_requires: Optional[List] = populate_dflts["install_requires"],
    long_description=populate_dflts["long_description"],
    long_description_content_type=populate_dflts["long_description_content_type"],
    include_pip_install_instruction_in_readme=True,
    verbose: bool = populate_dflts["verbose"],
    overwrite: List = (),
    defaults_from: Optional[str] = None,
    skip_docsrc_gen=False,
    skip_ci_def_gen=False,
    version_control_system=None,
    ci_def_path=None,
    ci_tpl_path=None,
    project_type=populate_dflts["project_type"],
    **configs,
):
    """Populate project directory root with useful packaging files, if they're missing.

    >>> from wads.populate import populate_pkg_dir
    >>> import os  # doctest: +SKIP
    >>> name = 'wads'  # doctest: +SKIP
    >>> pkg_dir = f'/D/Dropbox/dev/p3/proj/i/{name}'  # doctest: +SKIP
    >>> populate_pkg_dir(pkg_dir,  # doctest: +SKIP
    ...                  description='Tools for packaging',
    ...                  root_url=f'https://github.com/i2mint',
    ...                  author='OtoSense')

    :param pkg_dir: The relative or absolute path of the working directory. Defaults to '.'.
    :type pkg_dir: str, optional
    :param version: The desired version
    :param description: Short description of project
    :param root_url: Root url of the code repository (not the url of the project, but one level up that!)
    :param author: Author of the package
    :param license: License name for the package (should be recognized by pypi). Default is 'mit'
    :param description_file: File name containing a description of the project. Default is 'README.md'
    :param keywords: Keywords to include in pypi publication
    :param install_requires: The (pip install) names of of the packages required to install the package we're generating
    :param long_description: Text of the long description. Default is "file:README.md" (takes contents of README.md)
    :param long_description_content_type: How to parse the long_description. Default is "text/markdown"
    :param verbose: Set to True if you want to log extra information during the process. Defaults to False.
    :type verbose: bool, optional: Whether to print a lot of stuff as project is being populated.
    :param default_from: Name of field to look up in wads_configs to get defaults from,
        or 'user_input' to get it from user input.
    :param skip_docsrc_gen: Skip the generation of documentation stuff
    :param skip_ci_def_gen: Skip the generation of the CI stuff
    :param version_control_system: 'github' or 'gitlab' (will TRY to be resolved from root url if not given)
    :param ci_def_path: Path of the CI definition
    :param ci_tpl_path: Pater of the template definition
    :param configs: Extra configurations
    :return:

    """

    # If the pkg_dir is a github url, then we'll clone it and populate the
    # resulting folder
    if pkg_dir.startswith("https://github.com") or pkg_dir.startswith("git@github.com"):
        url = pkg_dir
        return populate_proj_from_url(url)

    args_defaults = dict()
    if defaults_from is not None:
        if defaults_from == "user_input":  # TODO: Implement!
            args_defaults = dict()  # ... and then fill with user input
            raise NotImplementedError("Not immplemented yet")  # TODO: Implement
        else:
            try:
                wads_configs = json.load(open(wads_configs_file))
                args_defaults = wads_configs[defaults_from]
            except KeyError:
                raise KeyError(
                    f"{wads_configs_file} json didn't have a {defaults_from} field"
                )

    if isinstance(overwrite, str):
        overwrite = {overwrite}
    else:
        overwrite = set(overwrite)

    _clog = mk_conditional_logger(condition=verbose, func=print)
    pkg_dir = os.path.abspath(os.path.expanduser(pkg_dir))
    assert os.path.isdir(pkg_dir), f"{pkg_dir} is not a directory"
    pkg_dir = ensure_no_slash_suffix(pkg_dir)
    name = os.path.basename(pkg_dir)
    pjoin = lambda *p: os.path.join(pkg_dir, *p)

    if name not in os.listdir(pkg_dir):
        f = pjoin(name)
        _clog(f"... making directory {pkg_dir}")
        os.mkdir(f)
    if "__init__.py" not in os.listdir(pjoin(name)):
        f = pjoin(name, "__init__.py")
        _clog(f"... making an empty {f}")
        with open(f, "w") as fp:
            fp.write("")

    # Note: Overkill since we just made those things...
    if name not in os.listdir(pkg_dir) or "__init__.py" not in os.listdir(pjoin(name)):
        raise RuntimeError(
            "You should have a {name}/{name}/__init__.py structure. You don't."
        )

    if os.path.isfile(pjoin("setup.cfg")):
        with open(pjoin("setup.cfg"), "r"):
            pass

    kwargs = dict(
        version=version,
        description=description,
        root_url=root_url,
        author=author,
        license=license,
        description_file=description_file,
        long_description=long_description,
        long_description_content_type=long_description_content_type,
        keywords=keywords,
        install_requires=install_requires,
    )
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    # configs = dict(name=name, **configs, **kwargs, **args_defaults)
    # configs = dict(name=name, **args_defaults, **configs, **kwargs)
    configs = dict(ChainMap(dict(name=name), kwargs, configs, args_defaults))

    kwargs["description-file"] = kwargs.pop("description_file", "")

    assert configs.get("name", name) == name, (
        f"There's a name conflict. pkg_dir tells me the name is {name}, "
        f"but configs tell me its {configs.get('name')}"
    )
    _ensure_url_from_url_root_and_name(configs)
    configs["display_name"] = configs.get("display_name", configs["name"])

    def should_update(resource_name):
        return (resource_name in overwrite) or (
            not os.path.isfile(pjoin(resource_name))
        )

    def save_txt_to_pkg(resource_name, content):
        target_path = pjoin(resource_name)
        assert not os.path.isfile(target_path), f"{target_path} exists already"
        _clog(f"... making a {resource_name}")
        with open(pjoin(resource_name), "wt") as fp:
            fp.write(content)

    if should_update(".gitignore"):
        shutil.copy(gitignore_tpl_path, pjoin(".gitignore"))

    if project_type == "app":
        if should_update("requirements.txt"):
            with open(pjoin("requirements.txt"), "w") as f:
                pass

    else:  # project_type == 'lib' or None
        if should_update("setup.py"):
            shutil.copy(setup_tpl_path, pjoin("setup.py"))
        if should_update("setup.cfg"):
            _clog("... making a 'setup.cfg'")
            if "pkg-dir" in configs:
                del configs["pkg-dir"]
            write_configs(pjoin(""), configs)

    if should_update("LICENSE"):
        _license_body = license_body(configs["license"])
        save_txt_to_pkg("LICENSE", _license_body)

    if should_update("README.md"):
        readme_text = gen_readme_text(name, configs.get("description"))
        if include_pip_install_instruction_in_readme:
            readme_text += f"\n\nTo install:\t```pip install {name}```\n"
        save_txt_to_pkg("README.md", readme_text)

    if project_type is None or project_type == "lib":
        if not skip_docsrc_gen:
            # TODO: Figure out epythet and wads relationship -- right now, there's a reflexive dependency
            from epythet.setup_docsrc import make_docsrc

            make_docsrc(pkg_dir, verbose=verbose)

    if not skip_ci_def_gen:
        root_url = root_url or _get_root_url_from_pkg_dir(pkg_dir)
        version_control_system = (
            version_control_system or _url_to_version_control_system(root_url)
        )
        ci_def_path, ci_tpl_path = _resolve_ci_def_and_tpl_path(
            ci_def_path, ci_tpl_path, pkg_dir, version_control_system, project_type
        )
        if should_update(ci_def_path):
            assert name in ci_def_path and name in _get_pkg_url_from_pkg_dir(
                pkg_dir
            ), f"The name wasn't found in both the ci_def_path AND the git url, so I'm going to be safe and do nothing"
            # TODO: Handle user_email better than this and make more visible
            user_email = kwargs.get("user_email", "thorwhalen1@gmail.com")
            _add_ci_def(ci_def_path, ci_tpl_path, root_url, name, _clog, user_email)

    return name


_unknown_version_control_system = object()


def _ensure_url_from_url_root_and_name(configs: dict):
    if "url" not in configs and "url_root" in configs and "name" in configs:
        if not configs["url_root"].endswith("/"):
            configs["urls_root"] += "/"
        configs["url"] = configs["urls_root"] + configs["name"]
    return configs


def _url_to_version_control_system(url):
    if "github.com" in url:
        return "github"
    elif "gitlab" in url:
        return "gitlab"
    else:
        return _unknown_version_control_system


def _resolve_ci_def_and_tpl_path(
    ci_def_path, ci_tpl_path, pkg_dir, version_control_system, project_type
):
    if ci_def_path is None:
        if version_control_system == "github":
            ci_def_path = os.path.join(pkg_dir, ".github/workflows/ci.yml")
        elif version_control_system == "gitlab":
            ci_def_path = os.path.join(pkg_dir, ".gitlab-ci.yml")
        else:
            raise ValueError(f"Unknown root url type: Neither github.com nor gitlab!")
    if ci_tpl_path is None:
        if version_control_system == "github":
            if project_type == "app":
                ci_tpl_path = github_ci_tpl_deploy_path
            else:  # project_type == 'lib', etc.
                print(f"project_type is {project_type}")
                ci_tpl_path = github_ci_tpl_publish_path
        elif version_control_system == "gitlab":
            ci_tpl_path = gitlab_ci_tpl_path
        else:
            raise ValueError(f"Unknown root url type: Neither github.com nor gitlab!")
    print(f"{ci_tpl_path=}")
    return ci_def_path, ci_tpl_path


def _add_ci_def(ci_def_path, ci_tpl_path, root_url, name, clog, user_email):
    clog(f"... making a {ci_def_path}")
    with open(ci_tpl_path) as f_in:
        ci_def = f_in.read()
        ci_def = ci_def.replace("#PROJECT_NAME#", name)
        hostname = urlparse(root_url).netloc
        ci_def = ci_def.replace("#GITLAB_HOSTNAME#", hostname)
        ci_def = ci_def.replace("#USER_EMAIL#", user_email)
        os.makedirs(os.path.dirname(ci_def_path), exist_ok=True)
        with open(ci_def_path, "w") as f_out:
            f_out.write(ci_def)


def _get_pkg_url_from_pkg_dir(pkg_dir):
    """Look in the .git of pkg_dir and get the project url for it.

    Note: If the url found is an ssh url, it will be converted to an https one.
    """
    import re

    pkg_dir = ensure_no_slash_suffix(pkg_dir)
    pkg_git_url = git(command="remote get-url origin", work_tree=pkg_dir)

    # Check if the URL is an SSH URL and convert it to HTTPS if needed
    ssh_match = re.match(r"git@(.*?):(.*?)(?:\.git)?$", pkg_git_url)
    if ssh_match:
        domain, repo = ssh_match.groups()
        pkg_git_url = f"https://{domain}/{repo}"

    return pkg_git_url


def _get_root_url_from_pkg_dir(pkg_dir):
    """Look in the .git of pkg_dir, get the url, and make a root_url from it"""
    pkg_git_url = _get_pkg_url_from_pkg_dir(pkg_dir)
    name = os.path.basename(pkg_dir)
    assert pkg_git_url.endswith(name) or pkg_git_url[:-1].endswith(name), (
        f"The pkg_git_url doesn't end with the pkg name ({name}), "
        f"so I won't try to guess. pkg_git_url is {pkg_git_url}. "
        f"For what ever you're doing, maybe there's a way to explicitly specify "
        f"the root url you're looking for?"
    )
    return pkg_git_url[: -len(name)]


def update_pack_and_setup_py(
    target_pkg_dir, copy_files=("setup.py", "wads/data/MANIFEST.in")
):
    """Just copy over setup.py and pack.py (moving the original to be prefixed by '_'"""
    copy_files = set(copy_files)
    target_pkg_dir = ensure_no_slash_suffix(target_pkg_dir)
    name = os.path.basename(target_pkg_dir)
    contents = os.listdir(target_pkg_dir)
    assert {"setup.py", name}.issubset(
        contents
    ), f"{target_pkg_dir} needs to have all three: {', '.join({'setup.py', name})}"

    pjoin = lambda *p: os.path.join(target_pkg_dir, *p)

    for resource_name in copy_files:
        print(f'... copying {resource_name} from {wads_join("")} to {target_pkg_dir}')
        shutil.move(src=pjoin(resource_name), dst=pjoin("_" + resource_name))
        shutil.copy(src=wads_join(resource_name), dst=pjoin(resource_name))


# --------------------------------------------------------------------------------------
# Extra: Populating a project from a github url
import contextlib
import subprocess
from functools import partial

DFLT_PROJ_ROOT_ENVVAR = "DFLT_PROJ_ROOT_ENVVAR"


def clog(*args, condition=True, log_func=print, **kwargs):
    if condition:
        return log_func(*args, **kwargs)


@contextlib.contextmanager
def cd(newdir, verbose=True):
    """Change your working directory, do stuff, and change back to the original"""
    _clog = partial(clog, condition=verbose, log_func=print)
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        _clog(f"cd {newdir}")
        yield
    finally:
        _clog(f"cd {prevdir}")
        os.chdir(prevdir)


# TODO: Use config2py to specify these, so user can configure them
name_for_url_root = {
    "https://github.com/i2mint": "i2mint",
    "https://github.com/thorwhalen": "thor",
}

# TODO: Use config2py to specify these, so user can configure them
proj_root_dir_for_name = {
    "i2mint": "i",
    "thor": "t",
}


def _mk_default_project_description(org_slash_proj: str) -> str:
    org, proj_name = org_slash_proj.split("/")
    return f"{proj_name} should say it all, no?"


def _get_org_slash_proj(repo: str) -> str:
    """Gets an org/proj_name string from a url (assuming it's at the end)

    >>> _get_org_slash_proj('https://github.com/thorwhalen/ut/')
    'thorwhalen/ut'
    """
    *_, org, proj_name = ensure_no_slash_suffix(repo).split("/")
    return f"{org}/{proj_name}"


def _to_git_https_url(git_url):
    """
    Converts gitURLs (namely SSH ones) to their HTTPS equivalent.

    :param git_url: A string containing the Git SSH URL.
    :return: A string containing the equivalent HTTPS URL.
    """
    if git_url.startswith("https://github.com"):
        return git_url
    elif git_url.startswith("http://github.com"):
        # return https equivalent
        return git_url.replace("http", "https")
    elif git_url.startswith("git@github.com"):
        stripped_url = git_url.replace("git@", "").replace(".git", "")
        formatted_url = stripped_url.replace(":", "/")
        https_url = f"https://{formatted_url}"
        return https_url
    elif git_url.startswith("github.com"):
        return f"https://{git_url}"
    else:
        owner, repo, *remainder = git_url.split("/")
        if not remainder:
            return f"https://github.com/{owner}/{repo}"
        else:
            raise ValueError(f"Cannot convert {git_url} to HTTPS URL")


def populate_proj_from_url(
    url,
    proj_rootdir=os.environ.get(DFLT_PROJ_ROOT_ENVVAR, None),
    description=None,
    license=populate_dflts["license"],
    **kwargs,
):
    """git clone a repository and set the resulting folder up for packaging."""
    verbose = kwargs.get("verbose", True)
    _clog = partial(clog, condition=verbose)

    _clog(f"Populating project for {url=}...")

    https_url = _to_git_https_url(url)
    https_url = ensure_no_slash_suffix(https_url)

    assert proj_rootdir, (
        "Your proj_rootdir was empty -- "
        "specify it or set the DFLT_PROJ_ROOT_ENVVAR envvar"
    )
    _clog(f"{proj_rootdir=}")

    root_url, proj_name = os.path.dirname(https_url), os.path.basename(https_url)
    if description is None:
        description = get_github_project_description(https_url)
    url_name = name_for_url_root.get(root_url, None)
    if url_name:
        _clog(f"url_name={url_name}")

    if url_name is not None and url_name in proj_root_dir_for_name:
        proj_rootdir = os.path.join(proj_rootdir, proj_root_dir_for_name[url_name])
    _clog(f"proj_rootdir={proj_rootdir}")

    with cd(proj_rootdir):
        _clog(f"cloning {url}...")
        subprocess.check_output(f"git clone {url}", shell=True).decode()
        _clog(f"populating package folder...")
        populate_pkg_dir(
            os.path.join(proj_rootdir, proj_name),
            defaults_from=url_name,
            description=description,
            license=license,
            **kwargs,
        )


def get_github_project_description(
    repo: str, default_factory=_mk_default_project_description
):
    """Get project description from github repository, or default if not found"""
    import requests

    org_slash_proj = _get_org_slash_proj(repo)
    api_url = f"https://api.github.com/repos/{org_slash_proj}"
    r = requests.get(api_url)
    if r.status_code == 200:
        description = r.json().get("description", None)
        if description:
            return description
        else:
            return default_factory(org_slash_proj)
    else:
        raise RuntimeError(
            f"Request response status for {api_url} wasn't 200. Was {r.status_code}"
        )


def main():
    import argh  # TODO: replace by argparse, or require argh in wads?

    argh.dispatch_command(populate_pkg_dir)


if __name__ == "__main__":
    main()



"""Generate Python Package with setup.cfg and setup.py configured to embed dependency wheels and
install them.
"""

import tempfile
from functools import partial
from pathlib import Path
from typing import List, Union
import os
import shutil
from distutils.command.sdist import sdist
from setuptools import setup
from setuptools.command.install import install as _install

from wads.generate_project_wheels import generate_project_wheels


def generate_package(
    module_path: Union[str, Path],
    install_requires: List[str],
    output_path: Union[str, Path],
    glob_pattern: Union[str, List[str]] = None,
    version: str = "1.0.0",
):
    """Generate Python module package including wheels for requirements linked to git repos

    :param module_path: Path to python folder or .py file to package
    :param install_requires: required package list like in setup.py or requirements.txt
    :param output_path: Folder to output package. Must not already exist.
    :param glob_pattern: pattern searched from module path, i.e. "*.pkl"
    :param version: semantic version, i.e. "1.0.0"
    :return:
    """
    module_path = Path(module_path)
    assert module_path.exists(), f'module does not exist: "{str(module_path)}"'
    assert not os.path.exists(
        output_path
    ), f'output path already exists: "{output_path}"'
    glob_pattern = [glob_pattern] if isinstance(glob_pattern, str) else glob_pattern

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)

        generate_module_folder(module_path, dst_path=temp_path)

        (temp_path / "setup.cfg").write_text(
            setup_cfg_template(
                install_requires,
                name=module_path.stem,
                version=version,
                glob_pattern=glob_pattern,
            )
        )
        (temp_path / "setup.py").write_text(setup_py)
        (temp_path / "MANIFEST.in").write_text(
            manifest_in_template(module_path.stem, glob_pattern)
        )

        shutil.copytree(temp_path, output_path)


def generate_module_folder(module_path: Path, dst_path: Path):
    if module_path.is_dir():
        shutil.copytree(module_path, dst_path / module_path.name)
    elif module_path.is_file() and module_path.suffix == ".py":
        module_folder = dst_path / module_path.stem
        module_folder.mkdir()
        shutil.copy(module_path, module_folder / "__init__.py")
    else:
        raise ValueError("module path must be a directory or .py file")


def manifest_in_template(module_name: str, glob_pattern: List[str]):
    manifest_in = "recursive-include dist *.whl"
    if glob_pattern:
        for gp in glob_pattern:
            manifest_in += f"\nrecursive-include {module_name} {gp}"
    return manifest_in


def setup_cfg_template(
    install_requires: List[str],
    name: str,
    version: str = "1.0.0",
    glob_pattern: List[str] = None,
) -> str:
    setup_cfg = f"""[metadata]
name = {name}
version = {version}
url = https://github.com/i2mint/wads
platforms = any
description_file = README.md
root_url = https://github.com/i2mint/
license = None
author = i2mint/wads
author_email = andie.phan@analog.com
description = {name}
long_description = file:README.md
long_description_content_type = text/markdown

[options]
packages = find:
include_package_data = True
zip_safe = False
install_requires = 
    wads
"""
    for r in install_requires:
        if r != "wads":
            setup_cfg += f"    {r}\n"

    if glob_pattern:
        setup_cfg += "\n[options.package_data]\n* =\n"
        for gp in glob_pattern:
            setup_cfg += f"    {gp}\n"

    return setup_cfg


class CustomSdistCommand(sdist):
    """Custom sdist command to include wheels in the distribution"""

    description = "create a source distribution tarball with embedded wheels"

    def _remove_wheels_from_install_requires(self, wheel_package_names):
        project_dir = os.getcwd()
        requirements_filepath = os.path.join(project_dir, "requirements.txt")
        setup_cfg_filepath = os.path.join(project_dir, "setup.cfg")
        if os.path.isfile(requirements_filepath):
            self._remove_lines(requirements_filepath, wheel_package_names)
        if os.path.isfile(setup_cfg_filepath):
            self._remove_lines(setup_cfg_filepath, wheel_package_names)

    def _remove_lines(self, filepath, exclude_lines):
        with open(filepath, "r") as file:
            lines = file.readlines()

        # Find the line to delete and remove it
        new_lines = []
        for line in lines:
            if line.strip() not in exclude_lines:
                new_lines.append(line)
            else:
                print(f"Removing {line.strip()} from {filepath}")

        # Write the new lines back to the file
        with open(filepath, "w") as file:
            file.writelines(new_lines)

    def _save_original_files(self):
        project_dir = os.getcwd()

        requirements_filepath = os.path.join(project_dir, "requirements.txt")
        setup_cfg_filepath = os.path.join(project_dir, "setup.cfg")
        if os.path.isfile(requirements_filepath):
            with open(requirements_filepath, "r") as f:
                self._requirements_txt = f.read()
        else:
            self._requirements_txt = None
        if os.path.isfile(setup_cfg_filepath):
            with open(setup_cfg_filepath, "r") as f:
                self._setup_cfg = f.read()
        else:
            self._setup_cfg = None

    def _restore_original_files(self):
        project_dir = os.getcwd()

        if self._requirements_txt:
            print("Restoring requirements.txt")
            requirements_filepath = os.path.join(project_dir, "requirements.txt")
            if os.path.isfile(requirements_filepath):
                with open(requirements_filepath, "w") as f:
                    f.write(self._requirements_txt)
        if self._setup_cfg:
            print("Restoring setup.cfg")
            setup_cfg_filepath = os.path.join(project_dir, "setup.cfg")
            if os.path.isfile(setup_cfg_filepath):
                with open(setup_cfg_filepath, "w") as f:
                    f.write(self._setup_cfg)

    def _generate_wheels(self):

        if self.dist_dir is None:
            self.dist_dir = "dist"
        if os.path.exists(self.dist_dir):
            print(f"Deleting {self.dist_dir}")
            shutil.rmtree(self.dist_dir)
        os.mkdir(self.dist_dir)

        # create a temporary directory to store the wheels
        project_dir = os.getcwd()
        wheel_generation_dir = os.path.join(project_dir, "wheels")
        if os.path.exists(wheel_generation_dir):
            print(f"Deleting {wheel_generation_dir}")

            shutil.rmtree(wheel_generation_dir)
        os.mkdir(wheel_generation_dir)
        print("Generating dependency wheels")
        if github_credentials := os.environ.get("GITHUB_CREDENTIALS", None):
            print("Using github credentials from environment variable")
        git_info = generate_project_wheels(
            project_dir, wheel_generation_dir, github_credentials=github_credentials
        )
        wheel_package_names = {g["name"] for g in git_info}
        # remove wheels from install_requires to skip pip dependency pre-check
        self._remove_wheels_from_install_requires(wheel_package_names)
        # copy the wheels into the package directory
        wheelhouse_dir = os.path.join(wheel_generation_dir, "wheelhouse")
        for wheel_file in os.listdir(wheelhouse_dir):
            if wheel_file.endswith(".whl"):
                copy_wheel_file = os.path.join(wheelhouse_dir, wheel_file)
                shutil.copy2(copy_wheel_file, self.dist_dir)
                print(f"Copying to {copy_wheel_file}")

        # cleanup the temporary directory
        shutil.rmtree(wheel_generation_dir)

    def initialize_options(self) -> None:
        super().initialize_options()
        # generating wheels here to modify setup.cfg before it's automatically copied
        self._save_original_files()
        try:
            self._generate_wheels()
        except Exception as e:
            self._restore_original_files()
            raise e

    def run(self):
        super().run()
        self._restore_original_files()


class CustomInstallCommand(_install):
    """Custom pip install command to install embedded wheels"""

    def run(self):
        temp_dir = os.getcwd()

        # Install wheels from the dist directory of the tarball
        if os.path.isdir(dist_dir := os.path.join(temp_dir, "dist")):
            for wheel in os.listdir(dist_dir):
                if wheel.endswith(".whl"):
                    print(f"Installing {wheel}")
                    self.spawn(
                        [
                            "pip",
                            "install",
                            os.path.join(dist_dir, wheel),
                            f"--find-links={dist_dir}",
                        ]
                    )

        # Run Normal Install
        _install.run(self)


embedded_wheel_setup = partial(
    setup, cmdclass={"install": CustomInstallCommand, "sdist": CustomSdistCommand}
)
setup_py = (
    "from wads.package_module import embedded_wheel_setup\nembedded_wheel_setup()"
)



"""Handle non python installs, depending on the system"""

# TODO: Instead of os.system, use subprocess.run?

from typing import Iterable, Callable, NewType, Union

InstallCommand = NewType("InstallCommand", str)
InstallCommandsStr = NewType("InstallCommandsStr", str)
InstallCommands = Union[InstallCommand, InstallCommandsStr, Iterable[InstallCommand]]


# TODO: Tried to make return_annotation InstallCommandsStr, but linter complained
def ensure_new_line_separated_strings(strings: InstallCommands, sep="\n") -> str:
    if isinstance(strings, str):
        return strings
    assert isinstance(strings, Iterable), f"Not an iterable: {strings}"
    return sep.join(strings)


# TODO: Tried to make return_annotation Iterable[InstallCommand], but linter complained
def ensure_iterable_of_strings(strings, sep="\n") -> Iterable[str]:
    if isinstance(strings, str):
        return strings.split(sep)
    assert isinstance(strings, Iterable), f"Not an iterable: {strings}"
    return strings


def make_install_instructions_for_readme(commands_for_os: dict) -> str:
    """

    :param commands_for_os: A dictionary containing the commands for each OS,
    where the keys are the OS names (as in ``os.name``) and the values are the
    commands, in the form of a string that can be passed to ``os.system`` to be
    executed.
    :return: A string that can included in the "How to install" section of a README.md,
    which will contain the instructions for the user to install the package,
    listing the commands for each OS.

    >>> make_install_instructions_for_readme({'Linux': 'pip install .'})
    '### Linux\\n\\n```\\npip install .\\n```\\n'

    """
    return "\n".join(
        f"### {os_name}\n\n"
        f"```\n"
        f"{ensure_new_line_separated_strings(commands)}\n"
        f"```\n"
        for os_name, commands in commands_for_os.items()
    )


def make_install_instructions_for_ci(commands_for_os: dict) -> str:
    """

    :param commands_for_os: A dictionary containing the commands for each OS,
    where the keys are the OS names (as in ``os.name``) and the values are the
    commands, in the form of a list of strings that can be passed to ``os.system`` to be
    executed.
    :return: A string that can included in a github actions ci.yml file,
    which will contain the os specific commands to install the package.

    >>> make_install_instructions_for_ci({'Linux': ['pip install .']})
    '- name: Install on Linux\\n  run: |\\n    pip install .\\n'

    """
    return "\n".join(
        f"- name: Install on {os_name}\n"
        f"  run: |\n"
        f"    {ensure_new_line_separated_strings(commands)}\n"
        for os_name, commands in commands_for_os.items()
    )


def warn_about_error(command, error_obj):
    from warnings import warn

    warn(f"Failed to run command: {command}.\nError: {error_obj}")
    return True


def raise_error(command, error_obj):
    raise RuntimeError(f"Failed to run command: {command}.\nError: {error_obj}")


def run_install_commands(
    commands_for_os: dict,
    on_error: Callable[[InstallCommand, Exception], None] = warn_about_error,
):
    """

    :param commands_for_os: A dictionary containing the commands for each OS,
    where the keys are the OS names (as in ``os.name``) and the values are the
    commands, in the form of a list of strings that can be passed to ``os.system`` to be
    executed.
    :param on_error: A function that will be called if an error occurs when running a
    command. The function should take two arguments: the command that failed to run,
    and the error object. The function should return a boolean indicating whether to
    continue running the commands or not.

    """
    import os

    os_name = os.name
    if os_name not in commands_for_os:
        raise ValueError(f"OS not supported: {os_name}")

    commands = commands_for_os[os_name]

    for command in commands:
        try:
            os.system(command)
        except Exception as e:
            if on_error(command, e):
                continue
            else:
                break
